/* The MIT License

   Copyright (C) 2024 Giulio Genovese

   Author: Giulio Genovese <giulio.genovese@gmail.com>

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
   THE SOFTWARE.

 */

#include <stdio.h>
#include <getopt.h>
#include <unistd.h>
#include <htslib/vcf.h>
#include <htslib/kstring.h>
#include <htslib/kseq.h>
#include "regidx.h" // cannot use htslib/regdix.h see http://github.com/samtools/htslib/pull/761
#include "mocha.h"
#include "bcftools.h"

#define MOCHAPHASE_VERSION "2024-11-05"

// retrieve allelic shift information from BCF record
// assumes little endian architecture
static inline int bcf_get_allelic_shift(const bcf_fmt_t *fmt, int8_t *as_arr, int nsmpl) {
    if (!fmt || fmt->n != 1) return 0;
    int i;

#define BRANCH(type_t, bcf_type_vector_end, bcf_type_missing)                                                          \
    {                                                                                                                  \
        type_t *p = (type_t *)fmt->p;                                                                                  \
        for (i = 0; i < nsmpl; i++, p++) {                                                                             \
            if (p[0] == bcf_type_vector_end || p[0] == bcf_type_missing)                                               \
                as_arr[i] = 0;                                                                                         \
            else                                                                                                       \
                as_arr[i] = p[0];                                                                                      \
        }                                                                                                              \
    }
    switch (fmt->type) {
    case BCF_BT_INT8:
        BRANCH(int8_t, bcf_int8_vector_end, bcf_int8_missing);
        break;
    case BCF_BT_INT16:
        BRANCH(int16_t, bcf_int16_vector_end, bcf_int16_missing);
        break;
    case BCF_BT_INT32:
        BRANCH(int32_t, bcf_int32_vector_end, bcf_int32_missing);
        break;
    default:
        error("Unexpected type %d\n", fmt->type);
    }
#undef BRANCH

    return 1;
}

// update phase information from BCF record
// 1 if higher number allele received from the mother
// -1 if higher number allele received from the father
// notice that it does not check for p[0] and p[1] are bcf_type_vector_end, so use with care
// assumes little endian architecture
static inline int bcf_update_genotype_phase(const bcf_fmt_t *fmt, int8_t *gt_phase_arr, int nsmpl) {
    if (!fmt || fmt->n != 2) return 0;
    int i;

#define BRANCH(type_t)                                                                                                 \
    {                                                                                                                  \
        type_t *p = (type_t *)fmt->p;                                                                                  \
        for (i = 0; i < nsmpl; i++, p += 2) {                                                                          \
            type_t gt0 = bcf_gt_allele(p[0]);                                                                          \
            type_t gt1 = bcf_gt_allele(p[1]);                                                                          \
            if ((gt_phase_arr[i] == 1 && gt0 > gt1) || (gt_phase_arr[i] == -1 && gt1 > gt0)) {                         \
                p[0] = bcf_gt_phased(gt1);                                                                             \
                p[1] = bcf_gt_phased(gt0);                                                                             \
            }                                                                                                          \
        }                                                                                                              \
    }
    switch (fmt->type) {
    case BCF_BT_INT8:
        BRANCH(int8_t);
        break;
    case BCF_BT_INT16:
        BRANCH(int16_t);
        break;
    case BCF_BT_INT32:
        BRANCH(int32_t);
        break;
    default:
        error("Unexpected type %d\n", fmt->type);
    }
#undef BRANCH

    return 1;
}

// Plugin initialization and description
const char *about(void) { return "Integrates phase information across sample duplicates.\n"; }

static const char *usage_text(void) {
    return "\n"
           "About: Integrates phase information across sample duplicates. (version " MOCHAPHASE_VERSION
           " http://github.com/freeseek/mocha)\n"
           "Usage: bcftools +mochaphase [options] -d <duplicates_file> <in.vcf.gz> [<in2.vcf.gz> ...]\n"
           "\n"
           "Plugin Options:\n"
           "        --vcfs-file <file>          File of list of VCFs files to use\n"
           "    -d, --duplicates <file>         Tab delimited file with list of duplicates\n"
           "        --flipped-regs <file>       Region/sample pairs where the representation is flipped\n"
           "    -c, --custom-tag <string>       Use this tag instead of the default AS\n"
           "        --no-version                Do not append version and command line to the header\n"
           "    -o, --output <file|dir>         Write output to a file [standard output] or directory [.] if multiple inputs\n"
           "    -O, --output-type u|b|v|z[0-9]  u/b: un/compressed BCF, v/z: un/compressed VCF, 0-9: compression level "
           "[v]\n"
           "    -r, --regions <region>          restrict to comma-separated list of regions\n"
           "    -R, --regions-file <file>       Restrict to regions listed in a file\n"
           "        --regions-overlap 0|1|2     Include if POS in the region (0), record overlaps (1), variant "
           "overlaps (2) [1]\n"
           "    -t, --targets [^]<region>       Similar to -r but streams rather than index-jumps. Exclude regions "
           "with \"^\" prefix\n"
           "    -T, --targets-file [^]<file>    Similar to -R but streams rather than index-jumps. Exclude regions "
           "with \"^\" prefix\n"
           "        --targets-overlap 0|1|2     Include if POS in the region (0), record overlaps (1), variant "
           "overlaps (2) [0]\n"
           "        --threads <int>             Number of extra output compression threads [0]\n"
           "    -s, --samples [^]<list>         Comma separated list of samples to include (or exclude with \"^\" "
           "prefix)\n"
           "    -S, --samples-file [^]<file>    File of samples to include (or exclude with \"^\" prefix)\n"
           "        --force-samples             Only warn about unknown subset samples\n"
           "        --suffix <string>           Append string to output VCF names when working with multiple VCFs [.phased]\n"
           "    -W, --write-index[=FMT]         Automatically index the output files [off]\n"
           "\n"
           "Example:\n"
           "    bcftools +mochaphase --duplicates dups.tsv file1.bcf file2.bcf\n"
           "\n";
}

// Function to read sample IDs from the text file
// returns the number of lines in the file
// smpl_ids will be an array with the list of all samples in the file
// idxs will be a shorter array with the location the cumulative number of samples on each line
static int read_duplicates(const char *fn, const bcf_srs_t *srs, int **hdr_ids, int **smpl_ids, int **idxs, int force_samples) {
    htsFile *fp = hts_open(fn, "r");
    if (fp == NULL) error("Could not open %s: %s\n", fn, strerror(errno));

    int m_hdr_ids = 0, m_smpl_ids = 0, n_ids = 0, m_idxs = 0, n_idxs = 0;
    kstring_t str = {0, 0, NULL};
    int moff = 0, *off = NULL, i, ncols;
    while (hts_getline(fp, KS_SEP_LINE, &str) > 0) {
        ncols = ksplit_core(str.s, '\t', &moff, &off);
        hts_expand(int, n_ids + ncols, m_hdr_ids, *hdr_ids);
        hts_expand(int, n_ids + ncols, m_smpl_ids, *smpl_ids);
        int j = 0;
        for (i = 0; i < ncols; i++) {
            int k, id = -1;
            for (k = 0; k < srs->nreaders; k++) {
                bcf_hdr_t *hdr = bcf_sr_get_header(srs, k);
                id = bcf_hdr_id2int(hdr, BCF_DT_SAMPLE, &str.s[off[i]]);
                if (id >= 0) break;
            }
            if (id < 0) {
                if (force_samples) continue;
                error("Cannot find sample %s in the headers\n", &str.s[off[i]]);
            }
            (*hdr_ids)[n_ids + j] = k;
            (*smpl_ids)[n_ids + j] = id;
            j++;
        }
        n_ids += j;
        hts_expand(int, n_idxs + 1, m_idxs, *idxs);
        (*idxs)[n_idxs] = n_ids;
        n_idxs++;
    }
    free(off);
    free(str.s);
    hts_close(fp);
    return n_idxs;
}

// adapted from Petr Danecek's implementation of regidx_parse_reg() in regidx.c by replacing *se with *se && !isspace(*se)
#define MAX_COOR_0 REGIDX_MAX   // CSI and hts_itr_query limit, 0-based
static int regidx_parse_reg2(const char *line, char **chr_beg, char **chr_end, uint32_t *beg, uint32_t *end, void *payload, void *usr)
{
    char *ss = (char*) line;
    while ( *ss && isspace(*ss) ) ss++;
    if ( !*ss ) return -1;      // skip blank lines
    if ( *ss=='#' ) return -1;  // skip comments

    char *se = ss;
    while ( *se && !isspace(*se) && *se!=':' ) se++;

    *chr_beg = ss;
    *chr_end = se-1;

    if ( !(*se && !isspace(*se)) )
    {
        *beg = 0;
        *end = MAX_COOR_0;
        return 0;
    }

    ss = se+1;
    *beg = strtod(ss, &se);
    if ( ss==se ) { fprintf(stderr,"Could not parse reg line: %s\n", line); return -2; }
    if ( *beg==0 ) { fprintf(stderr,"Could not parse reg line, expected 1-based coordinate: %s\n", line); return -2; }
    (*beg)--;

    if ( !se[0] || !se[1] )
        *end = se[0]=='-' ? MAX_COOR_0 : *beg;
    else
    {
        ss = se+1;
        *end = strtod(ss, &se);
        if ( ss==se ) *end = *beg;
        else if ( *end==0 ) { fprintf(stderr,"Could not parse reg line, expected 1-based coordinate: %s\n", line); return -2; }
        else (*end)--;
    }
    return 0;
}

static int flipped_regs_parse(const char *line, char **chr_beg, char **chr_end, uint32_t *beg, uint32_t *end, void *payload, void *usr) {
    int ret = regidx_parse_reg2(line, chr_beg, chr_end, beg, end, NULL, NULL);
    if (ret != 0) return ret;

    // Skip the region field that were parsed above
    char *ss = (char *)line;
    while (*ss && isspace(*ss)) ss++;
    int i, id = -1;
    for (i = 0; i < 1; i++) {
        while (*ss && !isspace(*ss)) ss++;
        if (!*ss) return -2; // wrong number of fields
        while (*ss && isspace(*ss)) ss++;
    }
    if (!*ss) return -2;

    // Parse the payload
    int64_t *dat = (int64_t *)payload;
    bcf_srs_t *srs = (bcf_srs_t *)usr;
    for (i = 0; i < srs->nreaders; i++) {
        bcf_hdr_t *hdr = bcf_sr_get_header(srs, i);
        // find which VCF contains that sample
        id = bcf_hdr_id2int(hdr, BCF_DT_SAMPLE, ss);
        if (id >= 0) break;
    }

    if (id < 0) return -1;
    *dat = ((int64_t)i << 32) + (int64_t)id;
    return 0;
}

// Plugin main processing loop
int run(int argc, char **argv) {
    const char *pathname = NULL;
    const char *duplicates_fname = NULL;
    const char *flipped_regs_fname = NULL;
    const char *custom_tag = "AS";
    const char *output_fname_dir = NULL;
    const char *custom_sfx = ".phased";
    int output_type = FT_VCF;
    int regions_overlap = 1;
    int targets_overlap = 0;
    int clevel = -1;
    int n_threads = 0;
    int record_cmd_line = 1;
    int write_index = 0;
    const char *targets_list = NULL;
    int targets_is_file = 0;
    const char *regions_list = NULL;
    int regions_is_file = 0;
    const char *sample_names = NULL;
    int sample_is_file = 0;
    int force_samples = 0;
    regidx_t *flipped_regs_idx = NULL;
    regitr_t *flipped_regs_itr = NULL;

    static struct option loptions[] = {
        {"vcfs-file", required_argument, NULL, 1},       {"duplicates", required_argument, NULL, 'd'},
        {"flipped-regs", required_argument, NULL, 2},
        {"custom-tag", required_argument, NULL, 'c'},    {"output", required_argument, NULL, 'o'},
        {"output-type", required_argument, NULL, 'O'},   {"threads", required_argument, NULL, 9},
        {"regions", required_argument, NULL, 'r'},       {"regions-file", required_argument, NULL, 'R'},
        {"regions-overlap", required_argument, NULL, 3}, {"targets", required_argument, NULL, 't'},
        {"targets-file", required_argument, NULL, 'T'},  {"targets-overlap", required_argument, NULL, 4},
        {"samples", required_argument, NULL, 's'},       {"samples-file", required_argument, NULL, 'S'},
        {"force-samples", no_argument, NULL, 5},         {"no-version", no_argument, NULL, 8},
        {"suffix", required_argument, NULL, 6},          {"write-index", optional_argument, NULL, 'W'},
        {0, 0, 0, 0}};
    int c;
    char *tmp;
    while ((c = getopt_long(argc, argv, "h?d:c:o:O:r:R:t:T:s:S:W::", loptions, NULL)) >= 0) {
        switch (c) {
        case 1:
            pathname = optarg;
            break;
        case 'd':
            duplicates_fname = optarg;
            break;
        case 2:
            flipped_regs_fname = optarg;
            break;
        case 'c':
            custom_tag = optarg;
            break;
        case 'o':
            output_fname_dir = optarg;
            break;
        case 'O':
            switch (optarg[0]) {
            case 'b':
                output_type = FT_BCF_GZ;
                break;
            case 'u':
                output_type = FT_BCF;
                break;
            case 'z':
                output_type = FT_VCF_GZ;
                break;
            case 'v':
                output_type = FT_VCF;
                break;
            default: {
                clevel = strtol(optarg, &tmp, 10);
                if (*tmp || clevel < 0 || clevel > 9) error("The output type \"%s\" not recognised\n", optarg);
            }
            }
            if (optarg[1]) {
                clevel = strtol(optarg + 1, &tmp, 10);
                if (*tmp || clevel < 0 || clevel > 9)
                    error("Could not parse argument: --compression-level %s\n", optarg + 1);
            }
            break;
        case 9:
            n_threads = strtol(optarg, &tmp, 0);
            if (*tmp) error("Could not parse argument: --threads %s\n", optarg);
            break;
        case 8:
            record_cmd_line = 0;
            break;
        case 'r':
            regions_list = optarg;
            break;
        case 'R':
            regions_list = optarg;
            regions_is_file = 1;
            break;
        case 3:
            if (!strcasecmp(optarg, "0"))
                regions_overlap = 0;
            else if (!strcasecmp(optarg, "1"))
                regions_overlap = 1;
            else if (!strcasecmp(optarg, "2"))
                regions_overlap = 2;
            else
                error("Could not parse: --regions-overlap %s\n", optarg);
            break;
        case 't':
            targets_list = optarg;
            break;
        case 'T':
            targets_list = optarg;
            targets_is_file = 1;
            break;
        case 4:
            if (!strcasecmp(optarg, "0"))
                targets_overlap = 0;
            else if (!strcasecmp(optarg, "1"))
                targets_overlap = 1;
            else if (!strcasecmp(optarg, "2"))
                targets_overlap = 2;
            else
                error("Could not parse: --targets-overlap %s\n", optarg);
            break;
        case 's':
            sample_names = optarg;
            break;
        case 'S':
            sample_names = optarg;
            sample_is_file = 1;
            break;
        case 5:
            force_samples = 1;
            break;
        case 6:
            custom_sfx = optarg;
            break;
        case 'W':
            if (!(write_index = write_index_parse(optarg))) error("Unsupported index format '%s'\n", optarg);
            break;
        case 'h':
        case '?':
        default:
            error("%s", usage_text());
            break;
        }
    }
    if (!duplicates_fname) error("Tab delimited file with duplicate sample IDs needs to be specified with --duplicates\n%s", usage_text());

    bcf_srs_t *srs = bcf_sr_init();
    if ( regions_list )
    {
        srs->require_index = 1;
        bcf_sr_set_opt(srs,BCF_SR_REGIONS_OVERLAP,regions_overlap);
        if ( bcf_sr_set_regions(srs, regions_list, regions_is_file)<0 ) error("Failed to read the regions: %s\n",regions_list);
    }
    if ( targets_list )
    {
        bcf_sr_set_opt(srs,BCF_SR_TARGETS_OVERLAP,targets_overlap);
        if ( bcf_sr_set_targets(srs, targets_list, targets_is_file, 0)<0 ) error("Failed to read the targets: %s\n",targets_list);
    }

    if (n_threads && bcf_sr_set_threads(srs, n_threads) < 0) error("Failed to create threads\n");

    int n_files;
    char *stdin_str = "-", **filenames = &stdin_str;
    if (pathname) {
        filenames = hts_readlines(pathname, &n_files);
        if (!filenames) error("Failed to read from file %s\n", pathname);
    } else if (argc == optind) {
        if ( !isatty(fileno((FILE *)stdin)) ) {
            n_files = 1;
            filenames = &stdin_str;  // reading from stdin
        }
        else { error("%s", usage_text()); }
    } else {
        n_files = argc - optind;
        filenames = argv + optind;
    }
    if (!n_files) error("%s", usage_text());

    if (n_files > 1) {
        bcf_sr_set_opt(srs, BCF_SR_REQUIRE_IDX);
        bcf_sr_set_opt(srs, BCF_SR_PAIR_LOGIC, BCF_SR_PAIR_EXACT);
    }

    // if option relative is active, generate structure for encoding haplotype information
    int i, j, k, l;
    int *gt_id = malloc(n_files * sizeof(int));
    int *as_id = malloc(n_files * sizeof(int));
    int8_t **phase_arr = (int8_t **)calloc(n_files, sizeof(int8_t *));
    int8_t **as_arr = (int8_t **)calloc(n_files, sizeof(int8_t *));
    bcf_hdr_t *hdr;
    for (i = 0; i < n_files; i++) {
        if (!bcf_sr_add_reader(srs, filenames[i]))
            error("Error opening %s: %s\n", filenames[i], bcf_sr_strerror(srs->errnum));
        hdr = bcf_sr_get_header(srs, i);
        int nsmpl = bcf_hdr_nsamples(hdr);
        gt_id[i] = bcf_hdr_id2int(hdr, BCF_DT_ID, "GT");
        as_id[i] = bcf_hdr_id2int(hdr, BCF_DT_ID, custom_tag);
        phase_arr[i] = malloc(nsmpl * sizeof(int8_t));
        as_arr[i] = malloc(nsmpl * sizeof(int8_t));
    }

    // read duplicates file
    int *hdr_ids = NULL, *smpl_ids = NULL, *idxs = NULL;
    int n_idxs = read_duplicates(duplicates_fname, srs, &hdr_ids, &smpl_ids, &idxs, force_samples);

    // read flipped regions file
    if (flipped_regs_fname) {
        flipped_regs_idx = regidx_init(flipped_regs_fname, flipped_regs_parse, NULL, sizeof(int64_t), (void *)srs);
        if (!flipped_regs_idx) error("Error: failed to read flipped region file: --flipped-regs %s\n", flipped_regs_fname);
        flipped_regs_itr = regitr_init(flipped_regs_idx);
    }

    // open output VCFs
    htsFile **out_fhs = (htsFile **)malloc(n_files * sizeof(htsFile *));
    char **index_fname = (char **)malloc(n_files * sizeof(char *));
    char *suffix = NULL;
    if ( output_type & FT_BCF ) suffix = ".bcf";
    else if ( output_type & FT_GZ ) suffix = ".vcf.gz";
    else suffix = ".vcf";
    kstring_t str = {0, 0, NULL};
    for (i = 0; i < n_files; i++) {
        str.l = 0;
        if (n_files == 1) {
            if (output_fname_dir) kputs(output_fname_dir, &str);
        } else {
            if (output_fname_dir)
                kputsn_(output_fname_dir, strlen(output_fname_dir), &str);
            else
                kputc_('.', &str);
            if ( str.s[str.l-1] != '/' ) kputc_('/', &str);
            char *ptr = strrchr(filenames[i], '/');
            if (ptr) ptr++; else ptr = filenames[i];
            int len = strlen(ptr);
            if ( len >= 4 && !strcasecmp(".bcf",ptr+len-4) ) len -= 4;
            else if ( len >= 4 && !strcasecmp(".vcf",ptr+len-4) ) len -= 4;
            else if ( len >= 7 && !strcasecmp(".vcf.gz",ptr+len-7) ) len -= 7;
            else if ( len >= 8 && !strcasecmp(".vcf.bgz",ptr+len-8) ) len -= 8;
            kputsn_(ptr, len, &str);
            kputsn_(custom_sfx, strlen(custom_sfx), &str);
            kputs(suffix, &str);
        }
        char wmode[8];
        set_wmode(wmode, output_type, str.s, clevel);
        out_fhs[i] = hts_open(str.s ? str.s : "-", wmode);
        if (out_fhs[i] == NULL) error("[%s] Error: cannot write to \"%s\": %s\n", __func__, str.s, strerror(errno));
        if (n_threads) hts_set_opt(out_fhs[i], HTS_OPT_THREAD_POOL, srs->p); // TODO revise this maybe
        hdr = bcf_sr_get_header(srs, i);
        if (record_cmd_line) bcf_hdr_append_version(hdr, argc, argv, "bcftools_plugin");
        if (bcf_hdr_write(out_fhs[i], hdr) < 0) error("Unable to write to output VCF file\n");
        if (init_index2(out_fhs[i], hdr, str.s, &index_fname[i], write_index) < 0)
            error("Error: failed to initialise index for %s\n", str.s);
    }
    free(str.s);

    // loop through all the lines in the VCFs
    while (bcf_sr_next_line(srs)) {
        const char *contig = NULL;
        int pos = -1;
        // read all phase information and allelic shift values
        for (i = 0; i < srs->nreaders; i++) {
            if (!bcf_sr_has_line(srs, i)) continue;
            hdr = bcf_sr_get_header(srs, i);
            int nsmpl = bcf_hdr_nsamples(hdr);
            bcf1_t *line = bcf_sr_get_line(srs, i);
            contig = bcf_hdr_id2name(hdr, line->rid);
            pos = line->pos + 1;
            bcf_fmt_t *gt_fmt = bcf_get_fmt_id(line, gt_id[i]);
            int ret = bcf_get_genotype_phase(gt_fmt, phase_arr[i], nsmpl);
            if (!ret) memset(phase_arr[i], 0, nsmpl * sizeof(int8_t));
            bcf_fmt_t *as_fmt = bcf_get_fmt_id(line, as_id[i]);
            ret = bcf_get_allelic_shift(as_fmt, as_arr[i], nsmpl);
            if (!ret) memset(as_arr[i], 0, nsmpl * sizeof(int8_t));
        }

        // checked whether I am dealing with a flipped region
        if (flipped_regs_idx) {
            regidx_overlap(flipped_regs_idx, contig, pos-1, pos, flipped_regs_itr);
            if (flipped_regs_itr) {
                while (regitr_overlap(flipped_regs_itr)) {
                    int64_t payload = regitr_payload(flipped_regs_itr, int64_t);
                    k = (int)(payload >> 32);
                    l = (int)payload;
                    as_arr[k][l] *= -1;
                }
            }
        }

        // loop through each duplicate group
        for (i = 0; i < n_idxs; i++) {
            int alt_over = 0; // number of times alternate allele is over represented
            int alt_under = 0; // number of times alternate allele is under represented
            // fix the phase by forcing the over represented allele to be the first allele
            for (j = i ? idxs[i - 1] : 0; j < idxs[i]; j++) {
                k = hdr_ids[j];
                l = smpl_ids[j];
                int as = as_arr[k][l];
                if (as > 0) alt_over++;
                if (as < 0) alt_under++;
            }
            // absolute phasing strategy:
            // here the idea is that, if the alternate allele is overrepresented across all duplicates
            // then we set the alternate allele as inherited by the mother
            int decision = alt_over > alt_under ? 1 : alt_over < alt_under ? -1 : 0;
            for (j = i ? idxs[i - 1] : 0; j < idxs[i]; j++) {
                k = hdr_ids[j];
                l = smpl_ids[j];
                int phase = phase_arr[k][l];
                if ((phase == 1 || phase == -1) && decision)
                    phase_arr[k][l] = decision;
            }
// if (alt_over + alt_under > 0) fprintf(stderr, "%d) OVER/UNDER=%d/%d\n", i, alt_over, alt_under);
        }
        // update all phase information
        for (i = 0; i < srs->nreaders; i++) {
            if (!bcf_sr_has_line(srs, i)) continue;
            hdr = bcf_sr_get_header(srs, i);
            int nsmpl = bcf_hdr_nsamples(hdr);
            bcf1_t *line = bcf_sr_get_line(srs, i);
            bcf_fmt_t *gt_fmt = bcf_get_fmt_id(line, gt_id[i]);
            bcf_update_genotype_phase(gt_fmt, phase_arr[i], nsmpl);
            if (bcf_write(out_fhs[i], hdr, line) < 0) error("Unable to write to output VCF file\n");
        }
    }

    // close output VCFs
    for (i = 0; i < n_files; i++) {
        if (write_index) {
            if (bcf_idx_save(out_fhs[i]) < 0) {
                if (hts_close(out_fhs[i]) != 0) // output_fname_dir cannot be used like  this
                    error("Close failed %s\n", strcmp(output_fname_dir, "-") ? output_fname_dir : "stdout");
                error("Error: cannot write to index %s\n", index_fname[i]);
            }
            free(index_fname[i]);
        }
        hts_close(out_fhs[i]);
    }
    free(out_fhs);
    free(index_fname);

    free(gt_id);
    free(as_id);
    for (i = 0; i < n_files; i++) {
        free(phase_arr[i]);
        free(as_arr[i]);
    }
    free(phase_arr);
    free(as_arr);
    free(hdr_ids);
    free(smpl_ids);
    free(idxs);
    if (flipped_regs_idx) regidx_destroy(flipped_regs_idx);
    if (flipped_regs_itr) regitr_destroy(flipped_regs_itr);
    bcf_sr_destroy(srs);
    return 0;
}